import torch
from aspire.flows.torch.flows import ZukoFlow
from aspire.transforms import FlowTransform


def test_zuko_flow():
    dims = 3
    parameters = [f"x_{i}" for i in range(dims)]

    data_transform = FlowTransform(parameters=parameters, xp=torch)

    # Create an instance of ZukoFlow
    flow = ZukoFlow(
        dims=dims,
        flow_class="MAF",
        seed=42,
        device="cpu",
        data_transform=data_transform,
    )

    x = torch.randn(100, dims, device=flow.device)

    flow.fit_data_transform(x)

    # Check if the flow is initialized correctly
    assert flow.dims == dims

    # Check if the flow is an instance of ZukoFlow
    assert isinstance(flow, ZukoFlow)

    # Check if the flow has a valid flow attribute
    assert flow.flow is not None

    x = torch.tensor([0.1, 0.2, 0.3], device=flow.device)

    log_prob = flow.log_prob(x)

    assert log_prob.shape == (1,)



import jax
import jax.numpy as jnp
from aspire.flows.jax.flows import FlowJax
from aspire.transforms import FlowTransform


def test_zuko_flow():
    dims = 3
    parameters = [f"x_{i}" for i in range(dims)]

    data_transform = FlowTransform(parameters=parameters, xp=jnp)
    key = jax.random.key(42)
    key, flow_key = jax.random.split(key)

    # Create an instance of FlowJax
    flow = FlowJax(
        dims=dims,
        key=flow_key,
        device="cpu",
        data_transform=data_transform,
    )

    key, samples_key = jax.random.split(key)
    x = jax.random.normal(samples_key, (100, dims))

    flow.fit_data_transform(x)

    # Check if the flow is initialized correctly
    assert flow.dims == dims

    # Check if the flow is an instance of ZukoFlow
    assert isinstance(flow, FlowJax)

    x = jnp.array([0.1, 0.2, 0.3])

    log_prob = flow.log_prob(x)

    assert log_prob.shape == (1,)

    key, sample_key = jax.random.split(key)
    x = flow.sample(1)
    assert x.shape == (1, dims)



import logging
import math
from typing import Any

from array_api_compat import device as get_device
from array_api_compat import is_torch_namespace

from .flows import get_flow_wrapper
from .utils import (
    copy_array,
    logit,
    sigmoid,
    update_at_indices,
)

logger = logging.getLogger(__name__)


class BaseTransform:
    """Base class for data transforms.

    Parameters
    ----------
    xp : Callable
        The array API namespace to use (e.g., numpy, torch).
    dtype : Any, optional
        The data type to use for the transform. If not provided, defaults to
        the default dtype of the array API namespace if available.
    """

    def __init__(self, xp, dtype=None):
        self.xp = xp
        if is_torch_namespace(self.xp) and dtype is None:
            dtype = self.xp.get_default_dtype()
        self.dtype = dtype

    def fit(self, x):
        """Fit the transform to the data."""
        raise NotImplementedError("Subclasses must implement fit method.")

    def forward(self, x):
        raise NotImplementedError("Subclasses must implement forward method.")

    def inverse(self, y):
        raise NotImplementedError("Subclasses must implement inverse method.")


class IdentityTransform(BaseTransform):
    """Identity transform that does nothing to the data."""

    def fit(self, x):
        return copy_array(x, xp=self.xp)

    def forward(self, x):
        return copy_array(x, xp=self.xp), self.xp.zeros(
            len(x), device=get_device(x)
        )

    def inverse(self, y):
        return copy_array(y, xp=self.xp), self.xp.zeros(
            len(y), device=get_device(y)
        )


class CompositeTransform(BaseTransform):
    def __init__(
        self,
        parameters: list[int],
        periodic_parameters: list[int] = None,
        prior_bounds: list[tuple[float, float]] = None,
        bounded_to_unbounded: bool = True,
        bounded_transform: str = "probit",
        affine_transform: bool = True,
        device=None,
        xp: None = None,
        eps: float = 1e-6,
        dtype: Any = None,
    ):
        super().__init__(xp=xp, dtype=dtype)
        if prior_bounds is None:
            logger.warning(
                "Missing prior bounds, some transforms may not be applied."
            )
        if periodic_parameters and not prior_bounds:
            raise ValueError(
                "Must specify prior bounds to use periodic parameters."
            )
        self.parameters = parameters
        self.periodic_parameters = periodic_parameters or []
        self.bounded_to_unbounded = bounded_to_unbounded
        self.bounded_transform = bounded_transform
        self.affine_transform = affine_transform

        self.eps = eps
        self.device = device

        if prior_bounds is None:
            self.prior_bounds = None
            self.bounded_parameters = None
            lower_bounds = None
            upper_bounds = None
        else:
            logger.info(f"Prior bounds: {prior_bounds}")
            self.prior_bounds = {
                k: self.xp.asarray(
                    prior_bounds[k], device=device, dtype=self.dtype
                )
                for k in self.parameters
            }
            if bounded_to_unbounded:
                self.bounded_parameters = [
                    p
                    for p in parameters
                    if self.xp.isfinite(self.prior_bounds[p]).all()
                    and p not in self.periodic_parameters
                ]
            else:
                self.bounded_parameters = None
            lower_bounds = self.xp.asarray(
                [self.prior_bounds[p][0] for p in parameters],
                device=device,
                dtype=self.dtype,
            )
            upper_bounds = self.xp.asarray(
                [self.prior_bounds[p][1] for p in parameters],
                device=device,
                dtype=self.dtype,
            )

        if self.periodic_parameters:
            logger.info(f"Periodic parameters: {self.periodic_parameters}")
            self.periodic_mask = self.xp.asarray(
                [p in self.periodic_parameters for p in parameters],
                dtype=bool,
                device=device,
            )
            self._periodic_transform = PeriodicTransform(
                lower=lower_bounds[self.periodic_mask],
                upper=upper_bounds[self.periodic_mask],
                xp=self.xp,
            )
        if self.bounded_parameters:
            logger.info(f"Bounded parameters: {self.bounded_parameters}")
            self.bounded_mask = self.xp.asarray(
                [p in self.bounded_parameters for p in parameters], dtype=bool
            )
            if self.bounded_transform == "probit":
                BoundedClass = ProbitTransform
            elif self.bounded_transform == "logit":
                BoundedClass = LogitTransform
            else:
                raise ValueError(
                    f"Unknown bounded transform: {self.bounded_transform}"
                )

            self._bounded_transform = BoundedClass(
                lower=lower_bounds[self.bounded_mask],
                upper=upper_bounds[self.bounded_mask],
                xp=self.xp,
                eps=self.eps,
            )

        if self.affine_transform:
            logger.info(f"Affine transform applied to: {self.parameters}")
            self._affine_transform = AffineTransform(xp=self.xp)
        else:
            self._affine_transform = None

    def fit(self, x):
        x = copy_array(x, xp=self.xp)
        if self.periodic_parameters:
            logger.debug(
                f"Fitting periodic transform to parameters: {self.periodic_parameters}"
            )
            x = update_at_indices(
                x,
                (slice(None), self.periodic_mask),
                self._periodic_transform.fit(x[:, self.periodic_mask]),
            )
        if self.bounded_parameters:
            logger.debug(
                f"Fitting bounded transform to parameters: {self.bounded_parameters}"
            )
            x = update_at_indices(
                x,
                (slice(None), self.bounded_mask),
                self._bounded_transform.fit(x[:, self.bounded_mask]),
            )
        if self.affine_transform:
            logger.debug("Fitting affine transform to all parameters.")
            x = self._affine_transform.fit(x)
        return x

    def forward(self, x):
        x = copy_array(x, xp=self.xp)
        x = self.xp.atleast_2d(x)
        log_abs_det_jacobian = self.xp.zeros(len(x), device=self.device)
        if self.periodic_parameters:
            y, log_j_periodic = self._periodic_transform.forward(
                x[..., self.periodic_mask]
            )
            x = update_at_indices(x, (slice(None), self.periodic_mask), y)
            log_abs_det_jacobian += log_j_periodic

        if self.bounded_parameters:
            y, log_j_bounded = self._bounded_transform.forward(
                x[..., self.bounded_mask]
            )
            x = update_at_indices(x, (slice(None), self.bounded_mask), y)
            log_abs_det_jacobian += log_j_bounded

        if self.affine_transform:
            x, log_j_affine = self._affine_transform.forward(x)
            log_abs_det_jacobian += log_j_affine
        return x, log_abs_det_jacobian

    def inverse(self, x):
        x = copy_array(x, xp=self.xp)
        x = self.xp.atleast_2d(x)
        log_abs_det_jacobian = self.xp.zeros(len(x), device=self.device)
        if self.affine_transform:
            x, log_j_affine = self._affine_transform.inverse(x)
            log_abs_det_jacobian += log_j_affine

        if self.bounded_parameters:
            y, log_j_bounded = self._bounded_transform.inverse(
                x[..., self.bounded_mask]
            )
            x = update_at_indices(x, (slice(None), self.bounded_mask), y)
            log_abs_det_jacobian += log_j_bounded

        if self.periodic_parameters:
            y, log_j_periodic = self._periodic_transform.inverse(
                x[..., self.periodic_mask]
            )
            x = update_at_indices(x, (slice(None), self.periodic_mask), y)
            log_abs_det_jacobian += log_j_periodic

        return x, log_abs_det_jacobian

    def new_instance(self, xp=None):
        return self.__class__(
            parameters=self.parameters,
            periodic_parameters=self.periodic_parameters,
            prior_bounds=self.prior_bounds,
            bounded_to_unbounded=self.bounded_to_unbounded,
            bounded_transform=self.bounded_transform,
            device=self.device,
            xp=xp or self.xp,
            eps=self.eps,
        )


class FlowTransform(CompositeTransform):
    """Subclass of CompositeTransform that uses a Flow for transformations.

    Does not support periodic transforms.
    """

    def __init__(
        self,
        parameters: list[int],
        prior_bounds: list[tuple[float, float]] = None,
        bounded_to_unbounded: bool = True,
        bounded_transform: str = "probit",
        affine_transform: bool = True,
        device=None,
        xp=None,
        eps=1e-6,
        dtype=None,
    ):
        super().__init__(
            parameters=parameters,
            periodic_parameters=[],
            prior_bounds=prior_bounds,
            bounded_to_unbounded=bounded_to_unbounded,
            bounded_transform=bounded_transform,
            affine_transform=affine_transform,
            device=device,
            xp=xp,
            eps=eps,
            dtype=dtype,
        )

    def new_instance(self, xp=None):
        return self.__class__(
            parameters=self.parameters,
            prior_bounds=self.prior_bounds,
            bounded_to_unbounded=self.bounded_to_unbounded,
            bounded_transform=self.bounded_transform,
            device=self.device,
            xp=xp or self.xp,
            eps=self.eps,
        )


class PeriodicTransform(BaseTransform):
    name: str = "periodic"
    requires_prior_bounds: bool = True

    def __init__(self, lower, upper, xp, dtype=None):
        super().__init__(xp=xp, dtype=dtype)
        self.lower = xp.asarray(lower, dtype=dtype)
        self.upper = xp.asarray(upper, dtype=dtype)
        self._width = upper - lower
        self._shift = None

    def fit(self, x):
        return self.forward(x)[0]

    def forward(self, x):
        y = self.lower + (x - self.lower) % self._width
        return y, self.xp.zeros(y.shape[0], device=get_device(y))

    def inverse(self, y):
        x = self.lower + (y - self.lower) % self._width
        return x, self.xp.zeros(x.shape[0], device=get_device(x))


class ProbitTransform(BaseTransform):
    name: str = "probit"
    requires_prior_bounds: bool = True

    def __init__(self, lower, upper, xp, eps=1e-6, dtype=None):
        self.lower = xp.asarray(lower, dtype=dtype)
        self.upper = xp.asarray(upper, dtype=dtype)
        self._scale_log_abs_det_jacobian = -xp.log(upper - lower).sum()
        self.eps = eps
        self.xp = xp

    def fit(self, x):
        return self.forward(x)[0]

    def forward(self, x):
        from scipy.special import erfinv

        y = (x - self.lower) / (self.upper - self.lower)
        y = self.xp.clip(y, self.eps, 1.0 - self.eps)
        y = erfinv(2 * y - 1) * math.sqrt(2)
        log_abs_det_jacobian = (
            0.5 * (math.log(2 * math.pi) + y**2).sum(-1)
            + self._scale_log_abs_det_jacobian
        )
        return y, log_abs_det_jacobian

    def inverse(self, y):
        from scipy.special import erf

        log_abs_det_jacobian = (
            -(0.5 * (math.log(2 * math.pi) + y**2)).sum(-1)
            - self._scale_log_abs_det_jacobian
        )
        x = 0.5 * (1 + erf(y / math.sqrt(2)))
        x = (self.upper - self.lower) * x + self.lower
        return x, log_abs_det_jacobian


class LogitTransform(BaseTransform):
    name: str = "logit"
    requires_prior_bounds: bool = True

    def __init__(self, lower, upper, xp, eps=1e-6, dtype=None):
        self.lower = xp.asarray(lower, dtype=dtype)
        self.upper = xp.asarray(upper, dtype=dtype)
        self._scale_log_abs_det_jacobian = -xp.log(upper - lower).sum()
        self.eps = eps
        self.xp = xp

    def fit(self, x):
        return self.forward(x)[0]

    def forward(self, x):
        y = (x - self.lower) / (self.upper - self.lower)
        y, log_abs_det_jacobian = logit(y, eps=self.eps)
        log_abs_det_jacobian += self._scale_log_abs_det_jacobian
        return y, log_abs_det_jacobian

    def inverse(self, y):
        x, log_abs_det_jacobian = sigmoid(y)
        log_abs_det_jacobian -= self._scale_log_abs_det_jacobian
        x = (self.upper - self.lower) * x + self.lower
        return x, log_abs_det_jacobian


class AffineTransform(BaseTransform):
    name: str = "affine"
    requires_prior_bounds: bool = False

    def __init__(self, xp):
        self._mean = None
        self._std = None
        self.xp = xp

    def fit(self, x):
        self._mean = x.mean(0)
        self._std = x.std(0)
        self.log_abs_det_jacobian = -self.xp.log(self.xp.abs(self._std)).sum()
        return self.forward(x)[0]

    def forward(self, x):
        y = (x - self._mean) / self._std
        return y, self.log_abs_det_jacobian * self.xp.ones(
            y.shape[0], device=get_device(y)
        )

    def inverse(self, y):
        x = y * self._std + self._mean
        return x, -self.log_abs_det_jacobian * self.xp.ones(
            y.shape[0], device=get_device(y)
        )


class FlowPreconditioningTransform(BaseTransform):
    def __init__(
        self,
        parameters: list[int],
        flow_backend: str = "zuko",
        prior_bounds: list[tuple[float, float]] = None,
        bounded_to_unbounded: bool = True,
        bounded_transform: str = "probit",
        affine_transform: bool = True,
        periodic_parameters: list[int] = None,
        device=None,
        xp=None,
        eps=1e-6,
        dtype=None,
        flow_matching: bool = False,
        flow_kwargs: dict[str, Any] = None,
        fit_kwargs: dict[str, Any] = None,
    ):
        super().__init__(xp=xp, dtype=dtype)

        self.parameters = parameters
        self.periodic_parameters = periodic_parameters or []
        self.prior_bounds = prior_bounds
        self.bounded_to_unbounded = bounded_to_unbounded
        self.bounded_transform = bounded_transform
        self.affine_transform = affine_transform
        self.eps = eps
        self.device = device or "cpu"
        self.flow_backend = flow_backend
        self.flow_matching = flow_matching
        self.flow_kwargs = flow_kwargs or {}
        self.fit_kwargs = fit_kwargs or {}

        FlowClass = get_flow_wrapper(
            backend=flow_backend, flow_matching=flow_matching
        )
        transform = CompositeTransform(
            parameters=parameters,
            periodic_parameters=periodic_parameters,
            prior_bounds=prior_bounds,
            bounded_to_unbounded=bounded_to_unbounded,
            bounded_transform=bounded_transform,
            affine_transform=affine_transform,
            device=device,
            xp=FlowClass.xp,
            eps=eps,
            dtype=dtype,
        )

        self._data_transform = transform
        self._FlowClass = FlowClass
        self.flow = None

    def fit(self, x):
        self.flow = self._FlowClass(
            dims=len(self.parameters),
            device=self.device,
            data_transform=self._data_transform,
            **self.flow_kwargs,
        )
        self.flow.fit(x, **self.fit_kwargs)
        return self.flow.forward(x, xp=self.xp)[0]

    def forward(self, x):
        return self.flow.forward(x, xp=self.xp)

    def inverse(self, y):
        return self.flow.inverse(y, xp=self.xp)

    def new_instance(self, xp=None):
        return self.__class__(
            parameters=self.parameters,
            periodic_parameters=self.periodic_parameters,
            prior_bounds=self.prior_bounds,
            bounded_to_unbounded=self.bounded_to_unbounded,
            bounded_transform=self.bounded_transform,
            affine_transform=self.affine_transform,
            device=self.device,
            xp=xp or self.xp,
            eps=self.eps,
            dtype=self.dtype,
            flow_backend=self.flow_backend,
            flow_matching=self.flow_matching,
            flow_kwargs=self.flow_kwargs,
            fit_kwargs=self.fit_kwargs,
        )



import logging
from typing import Callable

import jax.numpy as jnp
import jax.random as jrandom
from flowjax.train import fit_to_data

from ..base import Flow
from .utils import get_flow

logger = logging.getLogger(__name__)


class FlowJax(Flow):
    xp = jnp

    def __init__(self, dims: int, key=None, data_transform=None, **kwargs):
        device = kwargs.pop("device", None)
        if device is not None:
            logger.warning("The device argument is not used in FlowJax. ")
        super().__init__(dims, device=device, data_transform=data_transform)
        if key is None:
            key = jrandom.key(0)
            logger.warning(
                "The key argument is None. "
                "A random key will be used for the flow. "
                "Results may not be reproducible."
            )
        self.key = key
        self.loc = None
        self.scale = None
        self.key, subkey = jrandom.split(self.key)
        self._flow = get_flow(
            key=subkey,
            dims=self.dims,
            **kwargs,
        )

    def fit(self, x, **kwargs):
        from ...history import FlowHistory

        x = jnp.asarray(x)
        x_prime = self.fit_data_transform(x)
        self.key, subkey = jrandom.split(self.key)
        self._flow, losses = fit_to_data(subkey, self._flow, x_prime, **kwargs)
        return FlowHistory(
            training_loss=list(losses["train"]),
            validation_loss=list(losses["val"]),
        )

    def forward(self, x, xp: Callable = jnp):
        x_prime, log_abs_det_jacobian = self.rescale(x)
        z, log_abs_det_jacobian_flow = self._flow.forward(x_prime)
        return xp.asarray(z), xp.asarray(
            log_abs_det_jacobian + log_abs_det_jacobian_flow
        )

    def inverse(self, z, xp: Callable = jnp):
        z = jnp.asarray(z)
        x_prime, log_abs_det_jacobian_flow = self._flow.inverse(z)
        x, log_abs_det_jacobian = self.inverse_rescale(x_prime)
        return xp.asarray(x), xp.asarray(
            log_abs_det_jacobian + log_abs_det_jacobian_flow
        )

    def log_prob(self, x, xp: Callable = jnp):
        x_prime, log_abs_det_jacobian = self.rescale(x)
        log_prob = self._flow.log_prob(x_prime)
        return xp.asarray(log_prob + log_abs_det_jacobian)

    def sample(self, n_samples: int, xp: Callable = jnp):
        self.key, subkey = jrandom.split(self.key)
        x_prime = self._flow.sample(subkey, (n_samples,))
        x = self.inverse_rescale(x_prime)[0]
        return xp.asarray(x)

    def sample_and_log_prob(self, n_samples: int, xp: Callable = jnp):
        self.key, subkey = jrandom.split(self.key)
        x_prime = self._flow.sample(subkey, (n_samples,))
        log_prob = self._flow.log_prob(x_prime)
        x, log_abs_det_jacobian = self.inverse_rescale(x_prime)
        return xp.asarray(x), xp.asarray(log_prob - log_abs_det_jacobian)



import logging
from typing import Any

from ..history import FlowHistory
from ..transforms import BaseTransform, IdentityTransform

logger = logging.getLogger(__name__)


class Flow:
    xp = None  # type: Any

    def __init__(
        self,
        dims: int,
        device: Any,
        data_transform: BaseTransform | None = None,
    ):
        self.dims = dims
        self.device = device

        if data_transform is None:
            data_transform = IdentityTransform(self.xp)
            logger.info("No data_transform provided, using IdentityTransform.")

        self.data_transform = data_transform

    def log_prob(self, x):
        raise NotImplementedError

    def sample(self, x):
        raise NotImplementedError

    def sample_and_log_prob(self, n_samples):
        raise NotImplementedError

    def fit(self, samples, **kwargs) -> FlowHistory:
        raise NotImplementedError

    def fit_data_transform(self, x):
        return self.data_transform.fit(x)

    def rescale(self, x):
        return self.data_transform.forward(x)

    def inverse_rescale(self, x):
        return self.data_transform.inverse(x)



import pytest
from aspire import Aspire


def test_integration_zuko(
    log_likelihood,
    log_prior,
    dims,
    samples,
    parameters,
    prior_bounds,
    bounded_to_unbounded,
    sampler_config,
):
    if sampler_config.sampler == "blackjax_smc":
        pytest.xfail(reason="BlackJAX requires JAX arrays.")

    aspire = Aspire(
        log_likelihood=log_likelihood,
        log_prior=log_prior,
        dims=dims,
        parameters=parameters,
        prior_bounds=prior_bounds,
        flow_matching=False,
        bounded_to_unbounded=bounded_to_unbounded,
        flow_backend="zuko",
    )
    aspire.fit(samples, n_epochs=5)
    aspire.sample_posterior(
        n_samples=100,
        sampler=sampler_config.sampler,
        **sampler_config.sampler_kwargs,
    )


@pytest.mark.requires("flowjax")
def test_integration_flowjax(
    log_likelihood,
    log_prior,
    dims,
    samples,
    parameters,
    prior_bounds,
    bounded_to_unbounded,
    samples_backend,
    sampler_config,
):
    import jax

    if sampler_config.sampler == "blackjax_smc" and samples_backend != "jax":
        pytest.xfail(reason="BlackJAX requires JAX arrays.")

    aspire = Aspire(
        log_likelihood=log_likelihood,
        log_prior=log_prior,
        dims=dims,
        parameters=parameters,
        prior_bounds=prior_bounds,
        flow_matching=False,
        bounded_to_unbounded=bounded_to_unbounded,
        flow_backend="flowjax",
        key=jax.random.key(0),
    )
    aspire.fit(samples, max_epochs=5)
    aspire.sample_posterior(
        n_samples=100,
        sampler=sampler_config.sampler,
        **sampler_config.sampler_kwargs,
    )



from __future__ import annotations

import inspect
import logging
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Any

import array_api_compat.numpy as np
import h5py
import wrapt
from array_api_compat import (
    array_namespace,
    is_jax_array,
    is_torch_array,
    is_torch_namespace,
    to_device,
)

if TYPE_CHECKING:
    from multiprocessing import Pool

    from array_api_compat.common._typing import Array

    from .aspire import Aspire

logger = logging.getLogger(__name__)


def configure_logger(
    log_level: str | int = "INFO",
    additional_loggers: list[str] = None,
    include_aspire_loggers: bool = True,
) -> logging.Logger:
    """Configure the logger.

    Adds a stream handler to the logger.

    Parameters
    ----------
    log_level : str or int, optional
        The log level to use. Defaults to "INFO".
    additional_loggers : list of str, optional
        Additional loggers to configure. Defaults to None.
    include_aspire_loggers : bool, optional
        Whether to include all loggers that start with "aspire_" or "aspire-".
        Defaults to True.

    Returns
    -------
    logging.Logger
        The configured logger.
    """
    logger = logging.getLogger("aspire")
    logger.setLevel(log_level)
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    formatter = logging.Formatter(
        "%(asctime)s - aspire - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    additional_loggers = additional_loggers or []
    for name in logger.manager.loggerDict:
        if include_aspire_loggers and (
            name.startswith("aspire_") or name.startswith("aspire-")
        ):
            additional_loggers.append(name)

    for name in additional_loggers:
        dep_logger = logging.getLogger(name)
        dep_logger.setLevel(log_level)
        dep_logger.handlers.clear()
        for handler in logger.handlers:
            dep_logger.addHandler(handler)
        dep_logger.propagate = False

    return logger


class PoolHandler:
    """Context manager to temporarily replace the log_likelihood method of a
    aspire instance with a version that uses a multiprocessing pool to
    parallelize computation.

    Parameters
    ----------
    aspire_instance : aspire
        The aspire instance to modify. The log_likelihood method of this
        instance must accept a :code:`map_fn` keyword argument.
    pool : multiprocessing.Pool
        The pool to use for parallel computation.
    close_pool : bool, optional
        Whether to close the pool when exiting the context manager.
        Defaults to True.
    parallelize_prior : bool, optional
        Whether to parallelize the log_prior method as well. Defaults to False.
        If True, the log_prior method of the aspire instance must also
        accept a :code:`map_fn` keyword argument.
    """

    def __init__(
        self,
        aspire_instance: Aspire,
        pool: Pool,
        close_pool: bool = True,
        parallelize_prior: bool = False,
    ):
        self.parallelize_prior = parallelize_prior
        self.aspire_instance = aspire_instance
        self.pool = pool
        self.close_pool = close_pool

    @property
    def aspire_instance(self):
        return self._aspire_instance

    @aspire_instance.setter
    def aspire_instance(self, value: Aspire):
        signature = inspect.signature(value.log_likelihood)
        if "map_fn" not in signature.parameters:
            raise ValueError(
                "The log_likelihood method of the Aspire instance must accept a"
                " 'map_fn' keyword argument."
            )
        signature = inspect.signature(value.log_prior)
        if "map_fn" not in signature.parameters and self.parallelize_prior:
            raise ValueError(
                "The log_prior method of the Aspire instance must accept a"
                " 'map_fn' keyword argument if parallelize_prior is True."
            )
        self._aspire_instance = value

    def __enter__(self):
        self.original_log_likelihood = self.aspire_instance.log_likelihood
        self.original_log_prior = self.aspire_instance.log_prior
        if self.pool is not None:
            logger.debug("Updating map function in log-likelihood method")
            self.aspire_instance.log_likelihood = partial(
                self.original_log_likelihood, map_fn=self.pool.map
            )
            if self.parallelize_prior:
                logger.debug("Updating map function in log-prior method")
                self.aspire_instance.log_prior = partial(
                    self.original_log_prior, map_fn=self.pool.map
                )
        return self.pool

    def __exit__(self, exc_type, exc_value, traceback):
        self.aspire_instance.log_likelihood = self.original_log_likelihood
        self.aspire_instance.log_prior = self.original_log_prior
        if self.close_pool:
            logger.debug("Closing pool")
            self.pool.close()
            self.pool.join()
        else:
            logger.debug("Not closing pool")


def logit(x: Array, eps: float | None = None) -> tuple[Array, Array]:
    """Logit function that also returns log Jacobian determinant.

    Parameters
    ----------
    x : float or ndarray
        Array of values
    eps : float, optional
        Epsilon value used to clamp inputs to [eps, 1 - eps]. If None, then
        inputs are not clamped.

    Returns
    -------
    float or ndarray
        Rescaled values.
    float or ndarray
        Log Jacobian determinant.
    """
    xp = array_namespace(x)
    if eps:
        x = xp.clip(x, eps, 1 - eps)
    y = xp.log(x) - xp.log1p(-x)
    log_j = (-xp.log(x) - xp.log1p(-x)).sum(-1)
    return y, log_j


def sigmoid(x: Array) -> tuple[Array, Array]:
    """Sigmoid function that also returns log Jacobian determinant.

    Parameters
    ----------
    x : float or ndarray
        Array of values

    Returns
    -------
    float or ndarray
        Rescaled values.
    float or ndarray
        Log Jacobian determinant.
    """
    xp = array_namespace(x)
    x = xp.divide(1, 1 + xp.exp(-x))
    log_j = (xp.log(x) + xp.log1p(-x)).sum(-1)
    return x, log_j


def logsumexp(x: Array, axis: int | None = None) -> Array:
    """Implementation of logsumexp that works with array api.

    This will be removed once the implementation in scipy is compatible.
    """
    xp = array_namespace(x)
    c = x.max()
    return c + xp.log(xp.sum(xp.exp(x - c), axis=axis))


def to_numpy(x: Array, **kwargs) -> np.ndarray:
    """Convert an array to a numpy array.

    This automatically moves the device to the CPU.

    Parameters
    ----------
    x : Array
        The array to convert.
    kwargs : dict
        Additional keyword arguments to pass to numpy.asarray.
    """
    try:
        return np.asarray(to_device(x, "cpu"), **kwargs)
    except ValueError:
        return np.asarray(x, **kwargs)


def asarray(x, xp: Any = None, **kwargs) -> Array:
    """Convert an array to the specified array API.

    Parameters
    ----------
    x : Array
        The array to convert.
    xp : Any
        The array API to use for the conversion. If None, the array API
        is inferred from the input array.
    kwargs : dict
        Additional keyword arguments to pass to xp.asarray.
    """
    if is_jax_array(x) and is_torch_namespace(xp):
        return xp.utils.dlpack.from_dlpack(x)
    else:
        return xp.asarray(x, **kwargs)


def copy_array(x, xp: Any = None) -> Array:
    """Copy an array based on the array API being used.

    This uses the most appropriate method to copy the array
    depending on the array API.

    Parameters
    ----------
    x : Array
        The array to copy.
    xp : Any
        The array API to use for the copy.

    Returns
    -------
    Array
        The copied array.
    """
    if xp is None:
        xp = array_namespace(x)
    # torch does not play nicely since it complains about copying tensors
    if is_torch_namespace(xp):
        if is_torch_array(x):
            return xp.clone(x)
        else:
            return xp.as_tensor(x)
    else:
        try:
            return xp.copy(x)
        except (AttributeError, TypeError):
            # Fallback for array APIs that do not have a copy method
            return xp.array(x, copy=True)


def effective_sample_size(log_w: Array) -> float:
    xp = array_namespace(log_w)
    return xp.exp(xp.asarray(logsumexp(log_w) * 2 - logsumexp(log_w * 2)))


@contextmanager
def disable_gradients(xp, inference: bool = True):
    """Disable gradients for a specific array API.

    Usage:

    ```python
    with disable_gradients(xp):
        # Do something
    ```

    Parameters
    ----------
    xp : module
        The array API module to use.
    inference : bool, optional
        When using PyTorch, set to True to enable inference mode.
    """
    if is_torch_namespace(xp):
        if inference:
            with xp.inference_mode():
                yield
        else:
            with xp.no_grad():
                yield
    else:
        yield


def encode_for_hdf5(value: Any) -> Any:
    """Encode a value for storage in an HDF5 file.

    Special cases:
    - None is replaced with "__none__"
    - Empty dictionaries are replaced with "__empty_dict__"
    """
    if isinstance(value, CallHistory):
        return value.to_dict(list_to_dict=True)
    if isinstance(value, np.ndarray):
        return value
    if isinstance(value, (int, float, str)):
        return value
    if isinstance(value, (list, tuple)):
        return [encode_for_hdf5(v) for v in value]
    if isinstance(value, set):
        return {encode_for_hdf5(v) for v in value}
    if isinstance(value, dict):
        if not value:
            return "__empty_dict__"
        else:
            return {k: encode_for_hdf5(v) for k, v in value.items()}
    if value is None:
        return "__none__"
    return value


def recursively_save_to_h5_file(h5_file, path, dictionary):
    """Recursively save a dictionary to an HDF5 file."""
    for key, value in dictionary.items():
        if isinstance(value, dict):
            recursively_save_to_h5_file(h5_file, f"{path}/{key}", value)
        else:
            try:
                h5_file.create_dataset(
                    f"{path}/{key}", data=encode_for_hdf5(value)
                )
            except TypeError as error:
                raise RuntimeError(
                    f"Cannot save key {key} with value {value} to HDF5 file."
                ) from error


def get_package_version(package_name: str) -> str:
    """Get the version of a package.

    Parameters
    ----------
    package_name : str
        The name of the package.

    Returns
    -------
    str
        The version of the package.
    """
    try:
        module = __import__(package_name)
        return module.__version__
    except ImportError:
        return "not installed"


class AspireFile(h5py.File):
    """A subclass of h5py.File that adds metadata to the file."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._set_aspire_metadata()

    def _set_aspire_metadata(self):
        from . import __version__ as aspire_version

        self.attrs["aspire_version"] = aspire_version


def update_at_indices(x: Array, slc: Array, y: Array) -> Array:
    """Update an array at specific indices."

    This is a workaround for the fact that array API does not support
    advanced indexing with all backends.

    Parameters
    ----------
    x : Array
        The array to update.
    slc : Array
        The indices to update.
    y : Array
        The values to set at the indices.

    Returns
    -------
    Array
        The updated array.
    """
    try:
        x[slc] = y
        return x
    except TypeError:
        return x.at[slc].set(y)


@dataclass
class CallHistory:
    """Class to store the history of calls to a function.

    Attributes
    ----------
    args : list[tuple]
        The positional arguments of each call.
    kwargs : list[dict]
        The keyword arguments of each call.
    """

    args: list[tuple]
    kwargs: list[dict]

    def to_dict(self, list_to_dict: bool = False) -> dict[str, Any]:
        """Convert the call history to a dictionary.

        Parameters
        ----------
        list_to_dict : bool
            If True, convert the lists of args and kwargs to dictionaries
            with string keys. If False, keep them as lists. This is useful
            when encoding the history for HDF5.
        """
        if list_to_dict:
            return {
                "args": {str(i): v for i, v in enumerate(self.args)},
                "kwargs": {str(i): v for i, v in enumerate(self.kwargs)},
            }
        else:
            return {
                "args": [list(arg) for arg in self.args],
                "kwargs": [dict(kwarg) for kwarg in self.kwargs],
            }


def track_calls(wrapped=None):
    """Decorator to track calls to a function.

    The decorator adds a :code:`calls` attribute to the wrapped function,
    which is a :py:class:`CallHistory` object that stores the arguments and
    keyword arguments of each call.
    """

    @wrapt.decorator
    def wrapper(wrapped_func, instance, args, kwargs):
        # If instance is provided, we're dealing with a method.
        if instance:
            # Attach `calls` attribute to the method's `__func__`, which is the original function
            if not hasattr(wrapped_func.__func__, "calls"):
                wrapped_func.__func__.calls = CallHistory([], [])
            wrapped_func.__func__.calls.args.append(args)
            wrapped_func.__func__.calls.kwargs.append(kwargs)
        else:
            # For standalone functions, attach `calls` directly to the function
            if not hasattr(wrapped_func, "calls"):
                wrapped_func.calls = CallHistory([], [])
            wrapped_func.calls.args.append(args)
            wrapped_func.calls.kwargs.append(kwargs)

        # Call the original wrapped function
        return wrapped_func(*args, **kwargs)

    return wrapper(wrapped) if wrapped else wrapper
